import json

from deprecated import deprecated
from ase import Atoms
from clease import basis_function as bf
from .concentration import Concentration
from .settings import ClusterExpansionSettings
from .settings_bulk import CEBulk, CECrystal
from .settings_slab import CESlab

__all__ = ('settings_from_json', 'settingsFromJSON')


def settings_from_json(fname):
    """Initialize settings from JSON.

    Exists due to compatibility. You should instead use
    `ClusterExpansionSettings.load(fname)`

    Parameters:

    fname: str
        JSON file where settings are stored
    """
    return ClusterExpansionSettings.load(fname)  # pylint: disable=no-member


def old_settings_from_json(fname):
    """
    Initialise settings from JSON file.
    Used for reading old json files from versions < 0.10.2

    Parameters:

    fname: str
        JSON file where settings are stored
    """
    with open(fname, 'r') as infile:
        data = json.load(infile)

    factory = data['kwargs'].pop('factory')
    kwargs = data['kwargs']
    conc = Concentration.from_dict(kwargs['concentration'])
    kwargs['concentration'] = conc
    if factory == 'CEBulk':
        settings = CEBulk(**kwargs)
    elif factory == 'CECrystal':
        settings = CECrystal(**kwargs)
    elif factory == 'CESlab':
        cnv_cell_dict = kwargs.pop('conventional_cell')
        cnv_cell = Atoms.fromdict(cnv_cell_dict)
        kwargs['conventional_cell'] = cnv_cell
        settings = CESlab(**kwargs)
    else:
        raise ValueError(f"Unknown factory {factory}")
    settings.include_background_atoms = data['include_background_atoms']
    settings.skew_threshold = data['skew_threshold']
    bf_dict = data['basis_func_type']
    name = bf_dict.pop('name')

    if name == 'polynmial':
        settings.basis_func_type = bf.Polynomial(**bf_dict)
    elif name == 'trigonometric':
        settings.basis_func_type = bf.Trigonometric(**bf_dict)
    elif name == 'binary_linear':
        settings.basis_func_type = bf.BinaryLinear(**bf_dict)
    return settings


@deprecated(version='0.10.0', reason='You should use "settings_from_json" instead')
def settingsFromJSON(fname):
    return settings_from_json(fname)



# pylint: disable=undefined-variable

from .system_changes import *
from .four_vector import *
from .figure import Figure

ADDITIONAL = ('system_changes',)
__all__ = system_changes.__all__ + four_vector.__all__ + figure.__all__ + ADDITIONAL



# pylint: disable=undefined-variable
from .mc_observer import MCObserver
from .sgc_observer import SGCObserver
from .corr_func_observer import CorrelationFunctionObserver
from .snapshot_observer import Snapshot
from .energy_evolution_obs import EnergyEvolution
from .site_order_parameter import SiteOrderParameter
from .lowest_energy_observer import LowestEnergyStructure
from .diffraction_observer import DiffractionObserver
from .energy_plot_updater import EnergyPlotUpdater
from .concentration_observer import ConcentrationObserver
from .multi_state_sgc_observer import MultiStateSGCConcObserver, SGCState
from .entropy_prod_rate import *
from .acceptance_rate import *

ADDITIONAL = ("MCObserver", "SGCObserver", "CorrelationFunctionObserver", "Snapshot",
              "EnergyEvolution", "LowestEnergyStructure", "DiffractionObserver",
              "EnergyPlotUpdater", "ConcentrationObserver", "SiteOrderParameter",
              "MultiStateSGCConcObserver", "SGCState")

__all__ = ADDITIONAL + (entropy_prod_rate.__all__ + acceptance_rate.__all__)



from itertools import product
import functools
from typing import List, Tuple, Dict, Set, Iterator, Iterable
import numpy as np
from ase import Atoms
from ase.geometry import wrap_positions
from clease.datastructures import FourVector, Figure
from .cluster import Cluster
from .cluster_fingerprint import ClusterFingerprint

__all__ = ('ClusterGenerator', 'SitesWithinCutoff')


class ClusterGenerator:
    """
    Class for generating cluster info that is independent of the
    template
    """

    def __init__(self, prim_cell: Atoms) -> None:
        self.with_cutoff = SitesWithinCutoff(self)
        self.prim = prim_cell

    @property
    def prim(self) -> Atoms:
        return self._prim

    @prim.setter
    def prim(self, other: Atoms) -> None:
        if not isinstance(other, Atoms):
            raise TypeError("prim must be an Atoms object")
        self._prim = other
        # Get the transposed cell as a contiguous array
        self.prim_cell_T = np.array(other.get_cell().T)
        self.prim_cell_invT = np.linalg.inv(self.prim_cell_T)

    def __eq__(self, other) -> bool:
        if not isinstance(other, ClusterGenerator):
            return NotImplemented

        return np.allclose(self.prim_cell_invT, other.prim_cell_invT) and self.prim == other.prim

    def eucledian_distance_vec(self, x1: FourVector, x2: FourVector) -> np.ndarray:
        """
        Eucledian distance between to FourVectors in cartesian coordinates.

        :param x1: First vector
        :param x2: Second vector
        """

        euc1 = x1.to_cartesian(self.prim, transposed_cell=self.prim_cell_T)
        euc2 = x2.to_cartesian(self.prim, transposed_cell=self.prim_cell_T)
        return euc2 - euc1

    def to_four_vector(self, cartesian: np.ndarray, sublattice: int = None) -> FourVector:
        """Translate a position in Cartesian coordinates to its FourVector"""
        if not cartesian.ndim == 1:
            # XXX: This could instead return a List[FourVector] if we pass a 2d
            # set of coordinates.
            raise ValueError('Can only translate 1 position at a time.')

        if sublattice is None:
            sublattice = self.get_lattice(cartesian)
        cartesian = cartesian - self.prim.positions[sublattice, :]

        ix, iy, iz = np.round(self.prim_cell_invT.dot(cartesian)).astype(int)
        return FourVector(ix, iy, iz, sublattice)

    def to_cartesian(self, *four_vectors: FourVector) -> np.ndarray:
        """Convert one or more FourVectors into their Cartesian coordinates."""
        pos = np.zeros((len(four_vectors), 3))
        for ii, fv in enumerate(four_vectors):
            pos[ii, :] = fv.to_cartesian(self.prim, transposed_cell=self.prim_cell_T)
        return pos

    def get_lattice(self, pos: np.ndarray) -> int:
        """
        Return the corresponding sublattice of a cartesian position

        :param pos: array of length 3 Cartesian position
        """
        shifts = self.prim.get_positions()
        reshaped = np.reshape(pos, (1, 3))
        wrapped = wrap_positions(reshaped, self.prim.get_cell())
        diff_sq = np.sum((wrapped[0, :] - shifts)**2, axis=1)
        return int(np.argmin(diff_sq))

    def eucledian_distance(self, x1: FourVector, x2: FourVector) -> float:
        d = self.eucledian_distance_vec(x1, x2)
        return np.linalg.norm(d, ord=2)

    @property
    def shortest_diag(self) -> float:
        shortest = None
        cellT = self.prim.get_cell().T
        for w in product([-1, 0, 1], repeat=3):
            if all(x == 0 for x in w):
                continue

            diag = cellT.dot(w)
            length: float = np.sqrt(np.sum(diag**2))
            if shortest is None or length < shortest:
                shortest = length
        assert shortest is not None
        return shortest

    @property
    def num_sub_lattices(self) -> int:
        return len(self.prim)

    def sites_within_cutoff(self, cutoff: float, x0: FourVector) -> Iterable[FourVector]:
        min_diag = self.shortest_diag
        max_int = int(cutoff / min_diag) + 1

        def filter_func(fv: FourVector) -> bool:
            d = self.eucledian_distance(x0, fv)
            return 1E-5 < d < cutoff

        all_sites = product(range(-max_int, max_int + 1), range(-max_int, max_int + 1),
                            range(-max_int, max_int + 1), range(self.num_sub_lattices))

        return filter(filter_func, (FourVector(*site) for site in all_sites))

    def get_fp(self, figure: Figure) -> ClusterFingerprint:
        """Generate finger print for a Figure object.

        Args:
            figure (Figure): The figure to find the finger print from.

        Returns:
            ClusterFingerprint: The finger print of the cluster.
        """
        X = figure.to_cartesian(self.prim, transposed_cell=self.prim_cell_T)
        return positions_to_fingerprint(X)

    def prepare_within_cutoff(self, cutoff: float,
                              lattice: int) -> Dict[FourVector, Set[FourVector]]:
        """Prepare all sites which are within the cutoff sphere. Note, this only prepares sites
        which are pair-wise within the cutoff, and does not consider the distance to the
        center-of-mass. This needs to be checked for the individual figure which is created from
        these sites."""

        within_cutoff = {}
        x0 = FourVector(0, 0, 0, lattice)
        sites = set(self.with_cutoff.get(cutoff, lattice))
        within_cutoff[x0] = sites
        for s in sites:
            nearby = set(s1 for s1 in sites if s1 != s and self.eucledian_distance(s1, s) <= cutoff)

            within_cutoff[s] = nearby
        return within_cutoff

    def figure_iterator(self, size: int, cutoff: float, ref_lattice: int) -> Iterator[Figure]:
        """Iterate all possible figures of a given size, cutoff radius, and reference lattice.
        All figures are guaranteed to have a radius smaller than the cutoff radius.

        Args:
            size (int): Size of the clusters to be generated, e.g. 2 for 2-body clusters.
            cutoff (float): Cutoff sphere from the center of mass of the cluster.
            ref_lattice (int, optional): Site which clusters are generated from. Any cluster
            always contains this site.

        Yields:
            Iterator[Figure]: Iterable with all acceptable figures.
        """

        def is_figure_ok(figure: Figure, cutoff: float) -> bool:
            """Does the figure obey the cutoff radius?
            The diameter can be larger than the maximum distance between two atoms
            in the cluster, and so the site_iterator can prepare clusters with too large
            diameters.

            Returns:
                bool: The diameter of the cluster is within the cutoff
            """
            return figure.get_diameter(self.prim, transposed_cell=self.prim_cell_T) < cutoff

        cutoff_lut = self.prepare_within_cutoff(cutoff, ref_lattice)
        # Function which filters figures based on the cutoff diameter from the center of mass
        filter_func = functools.partial(is_figure_ok, cutoff=cutoff)
        return filter(filter_func, site_iterator(cutoff_lut, size, ref_lattice))

    def generate(self, size: int, cutoff: float,
                 ref_lattice: int) -> Tuple[List[List[Figure]], List[ClusterFingerprint]]:
        """Generate all possible figures of a given size, are within a given cutoff radius
        (from the center of mass of the figure), and from a reference lattice.

        Args:
            size (int): Size of the clusters to be generated, e.g. 2 for 2-body clusters.
            cutoff (float): Cutoff sphere from the center of mass of the cluster.
            ref_lattice (int, optional): Site which clusters are generated from. Any cluster
            always contains this site.

        Returns:
            List[List[Figure]], List[ClusterFingerprint]: The collection of figures and
                their corresponding fingerprints.
        """
        clusters: List[List[Figure]] = []
        all_fps: List[ClusterFingerprint] = []

        for new_figure in self.figure_iterator(size, cutoff, ref_lattice):
            # The entries in the figure must be ordered, due to equiv_sites
            # TODO: This ordereding should not be necessary.
            new_figure = self._order_by_internal_distances(new_figure)

            fp = self.get_fp(new_figure)

            # Find the group
            try:
                group = all_fps.index(fp)
                clusters[group].append(new_figure)
            except ValueError:
                # Does not exist, create a new group
                clusters.append([new_figure])
                all_fps.append(fp)

        return clusters, all_fps

    def _get_internal_distances(self, figure: Figure) -> List[List[float]]:
        """
        Return all the internal distances of a figure
        """
        dists = []
        for x0 in figure.components:
            tmp_dist = []
            for x1 in figure.components:
                dist = self.eucledian_distance(x0, x1)
                # float() to ensure we're not working on a NumPy floating number
                dist = round(float(dist), 6)
                tmp_dist.append(dist)
            dists.append(sorted(tmp_dist, reverse=True))
        return dists

    def _order_by_internal_distances(self,
                                     figure: Figure) \
            -> Figure:
        """Order the Figure by internal distances. Returns a new instance of the Figure."""
        dists = self._get_internal_distances(figure)
        fvs = figure.components
        zipped = sorted(list(zip(dists, fvs)), reverse=True)
        return Figure((x[1] for x in zipped))

    def to_atom_index(self, cluster: Cluster, lut: Dict[FourVector, int]) -> List[List[int]]:
        """
        Convert the integer vector representation to an atomic index

        :param clusters: List of clusters (in integer vector representation)

        :param template: Template atoms to use

        :param lut: Look up table for 4-vectors to indices
        """
        # pylint: disable=no-self-use
        return [[lut[fv] for fv in fig.components] for fig in cluster.figures]

    def equivalent_sites(self, figure: Figure) -> List[List[int]]:
        """Find the equivalent sites of a figure."""
        dists = self._get_internal_distances(figure)
        equiv_sites = []
        for i, d1 in enumerate(dists):
            for j in range(i + 1, len(dists)):
                d2 = dists[j]
                if np.allclose(d1, d2):
                    equiv_sites.append((i, j))

        # Merge pairs into groups
        merged: List[Set[int]] = []
        for equiv in equiv_sites:
            found_group = False
            for m in merged:
                if any(x in m for x in equiv):
                    m.update(equiv)
                    found_group = True
            if not found_group:
                merged.append(set(equiv))
        return [list(x) for x in merged]

    def get_max_distance(self, figure: Figure):
        """Return the maximum distance of a figure."""
        internal_dists = self._get_internal_distances(figure)
        max_dists = [x[0] for x in internal_dists]
        return max(max_dists)


class SitesWithinCutoff:
    """
    Class that keeps track of nearby sites. It automatically re-calculates if
    the cutoff changes. If it can re-use a previous result it will do that.
    """

    def __init__(self, generator: ClusterGenerator) -> None:
        self.generator = generator
        self.cutoff = 0.0
        # Dictionary mapping a ref lattice to a list of four-vectors
        self.pre_calc: Dict[int, List[FourVector]] = {}

    def must_generate(self, cutoff: float, ref_lattice: int) -> bool:
        """
        Check if we can re-use a precalculated result
        """
        if abs(cutoff - self.cutoff) > 1e-10:
            return True
        return ref_lattice not in self.pre_calc.keys()

    def get(self, cutoff: float, ref_lattice: int) -> List[FourVector]:
        """
        Return sites within the cutoff
        """
        if self.must_generate(cutoff, ref_lattice):
            sites = self.generator.sites_within_cutoff(cutoff, FourVector(0, 0, 0, ref_lattice))
            self.pre_calc[ref_lattice] = list(sites)
            self.cutoff = cutoff
        return self.pre_calc[ref_lattice]


def site_iterator(within_cutoff: Dict[FourVector, Set[FourVector]], size: int,
                  ref_lattice: int) -> Iterator[Figure]:
    """
    Return an iterator of all combinations of sites within a cutoff

    :param within_cutoff: Dictionary returned by `prepare_within_cutoff`

    :param size: Cluster size

    :param ref_lattice: Reference lattice
    """
    x0 = FourVector(0, 0, 0, ref_lattice)

    def recursive_yield(remaining, current):
        if len(current) == size - 1:
            # This is the full Figure object
            components = [x0] + current
            fig = Figure(components)
            yield fig
        else:
            for next_item in remaining:
                # Avoid double counting
                if any(next_item <= v for v in current):
                    continue
                rem = set.intersection(remaining, within_cutoff[next_item])
                yield from recursive_yield(rem, current + [next_item])

    for v in within_cutoff[x0]:
        rem = within_cutoff[x0].intersection(within_cutoff[v])
        yield from recursive_yield(rem, [v])


def positions_to_fingerprint(X: np.ndarray) -> ClusterFingerprint:
    """
    Generate finger print given a position matrix
    """
    # Center around the center of mass
    X = np.array(X)
    com = np.mean(X, axis=0)
    X -= com
    X = X.dot(X.T)
    # The positions squared
    diag_positions = np.diagonal(X)

    N = X.shape[0]
    # Put all the upper triangle off-diagonals into a 1D array one after another,
    # [off_diag1, off_diag2, ...]
    off_diag_len = N * (N - 1) // 2
    off_diag = np.zeros(off_diag_len)
    offset = 0
    for i in range(1, N):
        diag = np.diagonal(X, offset=i)
        off_diag[offset:(len(diag) + offset)] = diag
        offset += len(diag)
    # Sort positions, such that the largest position is first
    diag_positions = np.sort(diag_positions)[::-1]
    off_diag = np.sort(off_diag)
    # Append the off-diagonal elements after the diagonal, into a 1D array
    inner = np.append(diag_positions, off_diag)
    return ClusterFingerprint(fp=inner)



from typing import Any, Iterable
from functools import total_ordering
import numpy as np
import attr

__all__ = ('ClusterFingerprint',)


def _fingerprint_converter(x: Any) -> Any:
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, Iterable):
        x = np.array(x)
    return x


@total_ordering
@attr.s(eq=False, order=False)
class ClusterFingerprint:
    """Container for a Cluster Fingerprint."""

    fp: np.ndarray = attr.ib(converter=_fingerprint_converter,
                             validator=attr.validators.instance_of(np.ndarray))
    tol: float = attr.ib(default=1e-9)

    def __lt__(self, other) -> bool:
        if not isinstance(other, ClusterFingerprint):
            return NotImplemented

        self_len = len(self.fp)
        other_len = len(other.fp)

        if self_len < other_len:
            return True
        if self_len > other_len:
            return False

        diff = self.fp - other.fp
        for value in diff:
            if value < -self.tol:
                return True
            if value > self.tol:
                return False

        return False

    def __eq__(self, other) -> bool:
        if not isinstance(other, ClusterFingerprint):
            return NotImplemented

        if len(self.fp) != len(other.fp):
            return False
        return np.allclose(self.fp, other.fp, atol=self.tol)

    def __getitem__(self, i):
        return self.fp[i]

    def __len__(self):
        return len(self.fp)

    @fp.validator
    def _validate_fp(self, attribute, value) -> None:
        """The Fingerprint must be a 1d array."""
        # pylint: disable=unused-argument, no-self-use
        if not value.ndim == 1:
            raise ValueError(f"Fingerprint must be a 1d array, got {value.ndim}")



from typing import NamedTuple, Sequence
import typing
import ase

__all__ = ('SystemChange', 'SystemChanges')


class SystemChange(NamedTuple):
    index: int
    old_symb: str
    new_symb: str
    name: str

    def apply_change(self, atoms: ase.Atoms) -> None:
        """Apply the change to the atoms object. The atoms object is mutated in place.
        Note, no check is performed to ensure that the 'old_symb' is correct.

        :param atoms: ASE Atoms object to be mutated.
        """
        atoms.symbols[self.index] = self.new_symb

    def undo_change(self, atoms: ase.Atoms) -> None:
        """Undo the change to the atoms object. The atoms object is mutated in place.
        Note, that no check is performed to ensure that the changed symbol is what is expected.

        :param atoms: ASE Atoms object to be mutated.
        """
        atoms.symbols[self.index] = self.old_symb


# Type alias for multiple SystemChange objects
SystemChanges = Sequence[SystemChange]


def get_inverted_changes(system_changes: SystemChanges) -> typing.Iterator[SystemChange]:
    """Invert system changes by doing old_symbs -> new_symbs."""
    yield from (SystemChange(index=change.index,
                             old_symb=change.new_symb,
                             new_symb=change.old_symb,
                             name=change.name) for change in system_changes)



from typing import List, Tuple

import numpy as np
from matplotlib import pyplot as plt

from clease.regression import LinearRegression
from clease.tools import aicc


class Sparsifier:
    """
    Sparsifier can be used to remove coefficicients that has very low values.
    """

    def __init__(self):
        self._aicc = []
        self._num_features = []

    def clear(self):
        """
        Clears all attributes
        """
        self._aicc = []
        self._num_features = []

    def sparsify(self, fitter: LinearRegression, X: np.ndarray,
                 y: np.ndarray) -> Tuple[List[int], np.ndarray]:
        """
        Remove the feature corresponding to the smallest coefficient
        iteratively. On each step, modified Afaike information
        crieterion is calculated (AICC). The features in the model
        with the smallest AICC is selected. This will favour models
        with fewer coefficients.

        Parameters:

        fitter:
            LinearRegression fitting scheme

        X:
            Design matrix of size (N x M) where N is the number of data points
            and M is the number of features

        y:
            Numpy array with the target values. The length must be N.

        Return:
            List with the index of the coefficient selected in the model with
            the lowest CV score and the coefficient array.

            Example:

            If the following is returned
            [0, 4, 8, 9], [0.1, 0.5, 0.8, -1.0]
            it means that feature no. 0, 4, 8 and 9 was selected. The
            coefficients corresponding to these features are returned in the
            second array. The coefficients of all other features in the model
            is zero.
        """
        self.clear()
        best_selection = None
        best_coeff = None
        best_afaike = None

        min_num_features = 2
        removed = []
        X_masked = X.copy()
        for _ in range(X.shape[1] - min_num_features + 1):
            mask = [j for j in range(X.shape[1]) if j not in removed]
            X_masked = X[:, mask]
            coeff = fitter.fit(X_masked, y)
            min_coeff = np.argmin(np.abs(coeff))
            removed.append(mask[min_coeff])

            pred = X_masked.dot(coeff)
            diff = pred - y
            mse = np.mean(diff**2)
            afaike_ic = aicc(mse, len(mask), len(y))

            self._aicc.append(afaike_ic)
            self._num_features.append(X.shape[1] - len(removed) + 1)

            if best_afaike is None or afaike_ic < best_afaike:
                best_coeff = coeff
                best_selection = mask
                best_afaike = afaike_ic

        return best_selection, best_coeff

    def plot(self):
        """
        Creates a plot of the CV score as function of the number of data points

        Returns: pyplot.Figure
            Figure object used to plot
        """
        if not self._aicc:
            raise RuntimeError("Nothing to plot. Call sparsify first.")

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self._num_features, self._aicc)
        ax.set_xlabel("Num. features")
        ax.set_ylabel("AICC")
        return fig



from typing import Sequence
import numpy as np
from clease.datastructures import SystemChange
from .bias_potential import BiasPotential


# pylint: disable=too-many-instance-attributes
class GaussianKernelBiasPotential(BiasPotential):
    """
    Bias potential represented by a sum of Gaussian kernels.

    Parameters:

        xmin: float
            Minimum coordinate value

        xmax: float
            Maximum coordinate value

        num_kernels: int
            Number of Gaussian kernels

        width: float
            Width of the kernel (=exp(-(x/width)**2)

        getter: MCObserver
            Observer that can return the coordinate value after a given MC step.
            There are two requirements for the observer in addition to being a
            subclass of MCObsers.
            1. It needs to support None as an argument for the system changes.
            In that case, the code expects that the current value of the
            coordinate is returned
            2. It needs to support peak=True/False keyword. If peak=True,
            the code expects to get the coordinate after the change,
            but the observer should not update its current value.
    """

    def __init__(self, xmin=0.0, xmax=1.0, num_kernels=10, width=0.1, getter=None):

        self.xmin = xmin
        self.xmax = xmax
        self.width = width
        self.getter = getter
        self.limit = 0.01

        self.pad = int(np.sqrt(-np.log(self.limit)) + 1)
        self.num_kernels = num_kernels + 2 * self.pad
        self.coeff = np.zeros(self.num_kernels)
        self.xmin_corrected = self.xmin - self.pad * width
        self.xmax_corrected = self.xmax + self.pad * width
        self.centers = np.linspace(self.xmin_corrected, self.xmax_corrected, self.num_kernels)
        self.dx = (self.xmax_corrected - self.xmin_corrected) / \
            (self.num_kernels-1)

    def get_index(self, x):
        return int((x - self.xmin_corrected) / self.dx)

    def get_x(self, indx):
        return self.xmin_corrected + self.dx * indx

    def inside_range(self, x):
        """
        Return the indices of the gaussians that are that has a contribution of
        more than 0.01 at position x

        Parameters
            x: float
                Position to be evaluated

        Returns: int, int
            lower index, upper index
        """
        delta = self.width * np.sqrt(-np.log(self.limit))
        upper = x + delta
        lower = x - delta
        i_upper = self.get_index(upper)
        i_lower = self.get_index(lower)
        assert i_lower >= 0
        assert i_upper < self.num_kernels
        return i_lower, i_upper

    def _gaussian(self, x, x0):
        return np.exp(-((x - x0) / self.width)**2)

    def evaluate(self, x):
        low, high = self.inside_range(x)
        w = self._gaussian(x, self.centers[low:high + 1])
        return np.sum(self.coeff[low:high + 1] * w)

    def __call__(self, system_changes: Sequence[SystemChange]):
        x = self.getter(system_changes, peak=True)
        return self.evaluate(x)

    def local_update(self, x, dE):
        """Increase the local energy at x by an amount dE."""
        low, high = self.inside_range(x)
        w = self._gaussian(x, self.centers[low:high + 1])

        # Update the coefficient in such a way that the value at
        # point x is increased by dE and that the L2 norm of the
        # change in coefficient is minimised
        self.coeff[low:high + 1] += dE * w / np.sum(w**2)

    def slope(self, x):
        """Evaluate slope."""
        low, high = self.inside_range(x)
        w = self._gaussian(x, self.centers[low:high + 1])
        w *= -(x - self.centers[low:high + 1]) / self.width**2
        return np.sum(self.coeff[low:high + 1] * w)

    def ensure_zero_slope(self, x):
        """Change the coefficients such that the slope is zero at x."""
        low, high = self.inside_range(x)
        c = self.centers[low:high + 1]
        w = self._gaussian(x, self.centers[low:high + 1])
        denom = (w * (x - c) / self.width**2)**2
        lamb = self.slope(x) / np.sum(denom)
        self.coeff[low:high + 1] += lamb * (x - c) * w / self.width**2

    def todict(self):
        return {
            'xmin': self.xmin,
            'xmax': self.xmax,
            'num_kernels': self.num_kernels,
            'width': self.width,
            'coeff': self.coeff.tolist(),
            'centers': self.centers.tolist(),
            'dx': self.dx,
            'xmin_corrected': self.xmin_corrected,
            'xmax_corrected': self.xmax_corrected,
            'pad': self.pad
        }

    def from_dict(self, data):
        self.xmin = data['xmin']
        self.xmax = data['xmax']
        self.num_kernels = data['num_kernels']
        self.width = data['width']
        self.coeff = np.array(data['coeff'])
        self.centers = np.array(data['centers'])
        self.dx = data['dx']
        self.xmin_corrected = data['xmin_corrected']
        self.xmax_corrected = data['xmax_corrected']
        self.pad = data.get('pad', 3)

    def calculate_from_scratch(self, atoms):
        x = self.getter.calculate_from_scratch(atoms)
        return self.evaluate(x)

    def zero(self):
        self.coeff[:] = 0.0

    def get_coeff(self):
        return self.coeff



from .clease import Clease, MovedIgnoredAtomError
from .clease_vol_dep import CleaseVolDep
from .util import attach_calculator, get_ce_energy

__all__ = ['Clease', 'MovedIgnoredAtomError', 'attach_calculator', 'get_ce_energy', 'CleaseVolDep']



from typing import Sequence, List, Set, Tuple
import random
from abc import abstractmethod, ABC
from ase import Atoms
from clease.datastructures import SystemChange
from clease.tools import flatten
from .constraints import MCConstraint
from .swap_move_index_tracker import SwapMoveIndexTracker

__all__ = ('TrialMoveGenerator', 'RandomFlip', 'RandomSwap', 'MixedSwapFlip', 'TooFewElementsError',
           'RandomFlipWithinBasis')

DEFAULT_MAX_ATTEMPTS = 10000


class TooFewElementsError(Exception):
    """
    Error that indicates that there are too few elements to perform a certain operation on
    an atoms object. An example is a Monte Carlo step at fixed concentrations which consists
    of swapping to atoms. If there are less than 2 different species in the Atoms object, it
    makes no sense to swap two atoms (all energies are equal)
    """


class TrialMoveGenerator(ABC):
    """
    Class for producing trial moves.

    :param max_attempts: Maximum number of attempts to try to find a move that passes
        the constraints. If not constraints are added, this has no effect.
    """

    def __init__(self, max_attempts: int = DEFAULT_MAX_ATTEMPTS):
        super().__init__()
        self._constraints = []
        self.max_attempts = max_attempts

    def initialize(self, atoms: Atoms) -> None:
        """
        Initialize the generator.

        :param atoms: Atoms object used in the simulation
        """

    def remove_constraints(self) -> None:
        """
        Remove all constraints
        """
        self._constraints = []

    @property
    def name(self):
        return type(self).__name__

    def add_constraint(self, cnst: MCConstraint):
        """
        Add a constraint to the generator

        :param cnst: Constraint that must be satisfied for all trial moves
        """
        self._constraints.append(cnst)

    def _change_ok(self, system_changes: Sequence[SystemChange]) -> bool:
        for cnst in self._constraints:
            if not cnst(system_changes):
                return False
        return True

    @abstractmethod
    def get_single_trial_move(self) -> Sequence[SystemChange]:
        """
        Return a single trial move, must be implemented in sub-classes
        """

    def get_trial_move(self) -> Sequence[SystemChange]:
        """
        Produce a trial move that is consistent with all cosntraints
        """
        for _ in range(self.max_attempts):
            move = self.get_single_trial_move()
            if self._change_ok(move):
                return move
        raise RuntimeError("Can not produce a valid trial move")

    def on_move_accepted(self, changes: Sequence[SystemChange]) -> None:
        """
        Callback that is called by Monte Carlo after each accepted move

         :param change: Seqeunce of trial moves performed
        """

    def on_move_rejected(self, changes: Sequence[SystemChange]) -> None:
        """
        Callback that is called after a move is rejected

        :param change: Seqeunce of trial moves performed
        """


class SingleTrialMoveGenerator(TrialMoveGenerator, ABC):
    """
    Interface class for generators that return only one type of trial moves
    """
    CHANGE_NAME = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.CHANGE_NAME is None:
            raise ValueError("Subclasses must set CHANGE_NAME")

    def name_matches(self, change: SystemChange) -> bool:
        """
        Return true of the name of the passed system change matches
        the CHANGE_NAME attribute.

        :param change: a system change
        """
        return self.CHANGE_NAME == change.name

    def made_changes(self, changes: Sequence[SystemChange]) -> List[SystemChange]:
        """
        Extract the subset system changes made by an instance of itself.
        This method can be overrided in sublcasses, but the default behavior is
        to extract the subset of changes where the name matches.
        """
        return [change for change in changes if self.name_matches(change)]


class RandomFlip(SingleTrialMoveGenerator):
    """
    Generate trial moves where the symbol at a given site is flipped

    :param symbols: Set with all symbols considered in a move
    :param atoms: Atoms object for the simulation
    :param indices: List with all indices that should be considered. If None, all indices are
        considered
    """
    CHANGE_NAME = 'flip_move'

    def __init__(self, symbols: Set[str], atoms: Atoms, indices: List[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.symbols = symbols
        self.atoms = atoms

        if indices is None:
            self.indices = list(range(len(self.atoms)))
        else:
            self.indices = indices

    def get_single_trial_move(self) -> List[SystemChange]:
        pos = random.choice(self.indices)
        old_symb = self.atoms[pos].symbol
        new_symb = random.choice([s for s in self.symbols if s != old_symb])
        return [
            SystemChange(index=pos, old_symb=old_symb, new_symb=new_symb, name=self.CHANGE_NAME)
        ]


class RandomSwap(SingleTrialMoveGenerator):
    """
    Produce random swaps

    :param atoms: Atoms object in the MC simulation
    :param indices: List with indices that can be chosen from. If None, all indices
        can be chosen.
    """
    CHANGE_NAME = 'swap_move'

    def __init__(self, atoms: Atoms, indices: List[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.tracker = SwapMoveIndexTracker()
        self.indices = indices
        self.initialize(atoms)

    def initialize(self, atoms: Atoms) -> None:
        """
        Initializes the atoms tracker
        """
        self.tracker.init_tracker(atoms)
        if self.indices:
            self.tracker.filter_indices(self.indices)

        if self.tracker.num_symbols < 2:
            raise TooFewElementsError("After filtering there are less than two symbol type left. "
                                      "Must have at least two.")

    def get_single_trial_move(self) -> List[SystemChange]:
        """
        Create a swap move
        """
        symb_a = random.choice(self.tracker.symbols)
        symb_b = random.choice([s for s in self.tracker.symbols if s != symb_a])
        rand_pos_a = self.tracker.get_random_indx_of_symbol(symb_a)
        rand_pos_b = self.tracker.get_random_indx_of_symbol(symb_b)
        return [
            SystemChange(index=rand_pos_a, old_symb=symb_a, new_symb=symb_b, name=self.CHANGE_NAME),
            SystemChange(index=rand_pos_b, old_symb=symb_b, new_symb=symb_a, name=self.CHANGE_NAME)
        ]

    def on_move_accepted(self, changes: Sequence[SystemChange]):
        self.tracker.update_swap_move(changes)


class MixedSwapFlip(TrialMoveGenerator):
    """
    Class for generating trial moves in a mixed ensemble. A subset of the
    sites should maintain a constant concentrations, and a subset should
    maintain constant chemical potential. Thus, for the subset of sites
    where the concentration should be fixed, swap moves are proposed and
    for the subset that should have constant chemical potentia, flip moves
    are probosed (e.g. switching symbol type on a site)

    :param atoms: Atoms object used in the simulation
    :param swap_indices: List of indices that constitue the sub-lattice
        that should have fixed concentration
    :param flip_indices: List of indices that constitute the sub-lattice
        that should have fixed chemical potential
    :param flip_symbols: List of possible symbols that can be substituted
        on the lattice with fixed chemical potential.
    :param flip_prob: Probability of returning a flip move. The probability
        of returning a swap move is then 1 - flip_prob.
    """

    def __init__(self,
                 atoms: Atoms,
                 swap_indices: Sequence[int],
                 flip_indices: Sequence[int],
                 flip_symbols: Sequence[str],
                 flip_prob: float = 0.5,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.swapper = RandomSwap(atoms, swap_indices)
        self.flipper = RandomFlip(flip_symbols, atoms, flip_indices)
        self.flip_prob = flip_prob
        self.initialize(atoms)

    @property
    def generators(self) -> Tuple[SingleTrialMoveGenerator]:
        return (self.flipper, self.swapper)

    @property
    def weights(self) -> Tuple[float]:
        """The probability weights for each generator"""
        return (self.flip_prob, 1.0 - self.flip_prob)

    def initialize(self, atoms: Atoms) -> None:
        """
        Initialize the trial move generator
        """
        for gen in self.generators:
            gen.initialize(atoms)

    def get_single_trial_move(self) -> Sequence[SystemChange]:
        """
        Produce a single trial move. Return a swap move with
        probability
        """
        gen = random.choices(self.generators, weights=self.weights, k=1)[0]
        return gen.get_single_trial_move()

    def on_move_accepted(self, changes: Sequence[SystemChange]):
        """
        Callback triggered when a move have been accepted.
        """
        for gen in self.generators:
            gen_changes = gen.made_changes(changes)
            if gen_changes:
                gen.on_move_accepted(gen_changes)

    def on_move_rejected(self, changes: Sequence[SystemChange]) -> None:
        """
        Callback triggered when a move have been accepted.
        """
        for gen in self.generators:
            gen_changes = gen.made_changes(changes)
            if gen_changes:
                gen.on_move_rejected(gen_changes)


class RandomFlipWithinBasis(SingleTrialMoveGenerator):
    """
    Produce trial moves consisting of flips within each basis. Each basis
    is defined by a list of indices.

    Args:
        symbols: Sequence allowed symbols in each basis
        atoms: Atoms object to be used in the simulation for which the trial
            moves are produced
        indices: Sequence of sets of indices where each set specify the indices
            of a basis. Note len(symbols) == len(indices)

    Example:

    Create a generator for a rocksalt structure with two basis

    >>> from ase.build import bulk
    >>> from clease.montecarlo import RandomFlipWithinBasis
    >>> atoms = bulk("LiO", crystalstructure="rocksalt", a=3.9)*(3, 3, 3)
    >>> basis1 = [a.index for a in atoms if a.symbol == "Li"]
    >>> basis2 = [a.index for a in atoms if a.symbol == "O"]
    >>> generator = RandomFlipWithinBasis([["Li", "X"], ["O", "V"]], atoms, [basis1, basis2])
    """

    CHANGE_NAME = 'flip_within_basis_move'

    def __init__(self,
                 symbols: Sequence[Sequence[str]],
                 atoms: Atoms,
                 indices: Sequence[Sequence[int]] = None,
                 **kwargs):
        super().__init__(**kwargs)

        if len(symbols) != len(indices):
            raise ValueError("Possible symbols must be specified for all basis.")

        if not _symbols_in_basis_are_unique(symbols):
            raise ValueError("Symbols within each basis must be unique")

        if not _index_occur_only_once(indices):
            raise ValueError("Each index can only occur in one basis and only once in each basis.")

        self._flippers = [RandomFlip(s, atoms, i) for s, i in zip(symbols, indices)]

    def get_single_trial_move(self) -> Sequence[SystemChange]:
        """
        Produce a trial move by choosing a random flipper
        """
        flipper = random.choice(self._flippers)
        return flipper.get_single_trial_move()


def _symbols_in_basis_are_unique(nested_seq: Sequence[Sequence[str]]) -> bool:
    """
    Check if all items in the sub-lists are unique.
    """
    for sub in nested_seq:
        if len(sub) != len(set(sub)):
            return False
    return True


def _index_occur_only_once(nested_seq: Sequence[Sequence[int]]) -> bool:
    flattened = flatten(nested_seq)
    return len(flattened) == len(set(flattened))



# pylint: disable=undefined-variable
from . import swap_move_index_tracker
from .mc_evaluator import *
from .barrier_models import *
from .kmc_events import *
from .kinetic_monte_carlo import *
from .trial_move_generator import *
from . import base
from .bias_potential import BiasPotential
from .montecarlo import Montecarlo
from .sgc_montecarlo import SGCMonteCarlo
from .gaussian_kernel_bias_potential import GaussianKernelBiasPotential
from .binned_bias_potential import BinnedBiasPotential
from .random_bias_with_memory import RandomBiasWithMemory
from .metadynamics_sampler import MetaDynamicsSampler

ADDITIONAL = ('Montecarlo', 'SGCMonteCarlo', 'MetaDynamicsSampler', 'BinnedBiasPotential',
              'BiasPotential', 'GaussianKernelBiasPotential', 'RandomBiasWithMemory', 'base',
              'swap_move_index_tracker')
__all__ = ADDITIONAL + (barrier_models.__all__ + kinetic_monte_carlo.__all__ + kmc_events.__all__ +
                        trial_move_generator.__all__ + mc_evaluator.__all__)
